import { Resource, validation as v, BaseResourceType } from '@modal-sh/yasumi';
import { DataSource, ResourceIdConfig } from '@modal-sh/yasumi/backend';
import { Database } from 'duckdb-async';
import assert from 'assert';

type ID = number;

export const AutoincrementIdConfig = {
  // TODO
  generationStrategy: async () => {},
  schema: v.number(),
  serialize: (v: unknown) => v?.toString() ?? '',
  deserialize: (v: string) => Number(v),
}

export class DuckDbDataSource<
  Schema extends v.BaseSchema = v.BaseSchema,
  CurrentName extends string = string,
  CurrentRouteName extends string = string,
  Data extends object = v.Output<Schema>,
> implements DataSource<Data, ID> {
  private resource?: Resource<BaseResourceType & {
    schema: Schema,
    name: CurrentName,
    routeName: CurrentRouteName,
  }>;

  db?: Database;

  constructor() {
    // noop
  }

  async initialize() {
    assert(typeof this.resource !== 'undefined');

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const idSchema = idConfig.schema as v.BaseSchema;

    this.db = await Database.create('test.db');
    const clause = `CREATE TABLE IF NOT EXISTS ${this.resource.state.routeName}`;
    const resourceSchema = this.resource.schema as unknown as v.ObjectSchema<any>;
    const tableSchema = Object.entries(resourceSchema.entries)
      .map(([columnName, columnDefRaw]) => {
        const columnDef = columnDefRaw as unknown as v.BaseSchema;
        return [columnName, columnDef.type].join(' ');
      })
      .join(',');
    const sql = `${clause} (${idAttr} ${idSchema.type}, ${tableSchema});`;
    const con = await this.db.connect();
    const stmt = await con.prepare(sql);
    await stmt.run();
  }

  prepareResource<Schema extends v.BaseSchema>(resource: Resource<BaseResourceType & {
    schema: Schema,
    name: CurrentName,
    routeName: CurrentRouteName,
  }>) {
    resource.dataSource = resource.dataSource ?? this;
    const originalResourceId = resource.id;
    resource.id = <NewIdAttr extends BaseResourceType['idAttr'], NewIdSchema extends BaseResourceType['idSchema']>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) => {
      originalResourceId(newIdAttr, params);
      return resource as Resource<BaseResourceType & {
        name: CurrentName,
        routeName: CurrentRouteName,
        schema: Schema,
        idAttr: NewIdAttr,
        idSchema: NewIdSchema,
      }>;
    };
    this.resource = resource as any;
  }

  async getMultiple() {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      SELECT * FROM ${this.resource.state.routeName};
    `);
    const data = await stmt.all();
    return data as Data[];
  }

  async newId() {
    const idConfig = this.resource?.state.shared.get('idConfig') as ResourceIdConfig<any>;
    assert(typeof idConfig !== 'undefined');

    const theNewId = await idConfig.generationStrategy(this);
    return theNewId as ID;
  }

  async create(data: Data) {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const { [idAttr]: _, ...effectiveData } = data as Record<string, unknown>;

    const clause = `INSERT INTO ${this.resource.state.routeName}`;
    const keys = Object.keys(effectiveData).join(',');
    const values = Object.values(effectiveData).map((d) => JSON.stringify(d).replace(/"/g, "'")).join(',');
    const sql = `${clause} (${keys}) VALUES (${values});`;
    const con = await this.db.connect();
    const stmt = await con.prepare(sql);
    await stmt.all();
    const newData = {
      ...effectiveData
    };
    return newData as Data;
  }

  async getTotalCount() {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      SELECT COUNT(*) FROM ?;
    `, this.resource.state.routeName);
    const [data] = await stmt.all();
    return data as unknown as number;
  }

  async getById(id: ID) {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      SELECT * FROM ? WHERE ? = ?;
    `, this.resource.state.routeName, idAttr, id);
    const data = await stmt.all();
    return data as Data;
  }

  async getSingle() {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      SELECT * FROM ? LIMIT 1;
    `, this.resource.state.routeName);
    const [data] = await stmt.all();
    return data as Data;
  }

  async delete(id: ID) {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      DELETE FROM ? WHERE ? = ?;
    `, this.resource.state.routeName, idAttr, id);
    await stmt.run();
  }

  async emplace(id: ID, data: Data) {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      INSERT OR REPLACE INTO ? (?, ...?) VALUES (...?);
    `, this.resource.state.routeName, id, Object.keys(data), Object.values(data));
    const [newData] = await stmt.all();
    // TODO check if created flag
    return [newData, false] as [Data, boolean];
  }

  async patch(id: ID, data: Partial<Data>) {
    assert(typeof this.db !== 'undefined');
    assert(typeof this.resource !== 'undefined');

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string');

    const con = await this.db.connect();
    const stmt = await con.prepare(`
      UPDATE ? SET ? = ? WHERE ? = ?;
    `,
      this.resource.state.routeName,
      Object.keys(data), Object.values(data),
      idAttr, id
    );
    const [newData] = await stmt.all();
    return newData as Data;
  }
}
